/* 
 * Copyright (C) 1999-2001 Joachim Wieland <joe@mcknight.de>
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA 02111, USA.
 */

#include <ctype.h>
#include "jftpgw.h"     /* this also includes config_header.h */

struct hostlist_t* hostlist_init() {
	struct hostlist_t* hl;
	hl = (struct hostlist_t*) malloc(sizeof(struct hostlist_t));
	enough_mem(hl);

	hl->next                     = (struct hostlist_t*) 0;
	hl->host.ip.ip               = -1;
	hl->host.ip.netmask          = -1;
	hl->host.name                = (char*) 0;

	return hl;
}


struct hostlist_t* hostlist_push(struct hostlist_t** hl,
				 char* name,
				 unsigned int ip,
				 unsigned int netmask) {

	struct hostlist_t* h;
	if (! hl) {
		/* this should NOT happen */
		log(3, "hostlist was NULL in hostlist_push()");
		return (struct hostlist_t*) 0;
	}
	h = *hl;
	if (! h) {
		/* new */
		h = hostlist_init();
		*hl = h;
	} else {
		while (h->next) {
			h = h->next;
		}
		h->next = hostlist_init();
		h = h->next;
	}

	/* fill with content */
	h->next = (struct hostlist_t*) 0;
	h->host.ip.ip           = ip;
	h->host.ip.netmask      = netmask;
	h->host.name            = name;

	/* return base address */
	return *hl;
}

struct hostlist_t* hostlist_clone(struct hostlist_t* hl) {
	struct hostlist_t* new;

	if ( ! hl ) {
		return (struct hostlist_t*) 0;
	}

	new = (struct hostlist_t*) malloc(sizeof(struct hostlist_t));

	new->host.ip = hl->host.ip;
	if (hl->host.name) {
		new->host.name = strdup(hl->host.name);
		enough_mem(new->host.name);
	} else {
		new->host.name = (char*) 0;
	}
	new->next = hostlist_clone(hl->next);
	return new;
}

struct hostlist_t* hostlist_ip_push(struct hostlist_t** hl,
				    unsigned int ip,
				    unsigned int netmask) {

	return hostlist_push(hl, (char*) 0, ip, netmask);
}

struct hostlist_t* hostlist_name_push(struct hostlist_t** hl,
				      char* name) {

	return hostlist_push(hl, name, -1, -1);
}


struct hostlist_t* config_parse_ips(struct hostlist_t** hl,
							const char* line) {
	const char* p = line;
	int invalid = 0;
	char ipbuf[16];

	struct hostlist_t* h_list = (struct hostlist_t*) 0;

	unsigned int ip, netmask;

	ip = -1;
	netmask = -1;

	if ( !line ) { *hl = (struct hostlist_t*) '\0'; };

	for (;;) {
		int i =0;
		while (p && *p) {
			/* set p to the next part */
			if (isdigit((int)*p)) {
				if (p == line) {
					break;
				}
				if (isspace((int)*(p-1))) {
					break;
				}
			}
			p++;
		}
		if (!p || !*p) {
			break;
		}
		/* p theoretically points to an ip */
		if (isdigit((int)*p)) {
			while ((isdigit((int)*p) || *p == '.') && *p != '/') {
				if (i < sizeof(ipbuf)-1) {
					ipbuf[i++] = *p;
				}
				p++;
			}
			ipbuf[i] = '\0';
			log(9, "Found ip %s", ipbuf);
			ip = inet_addr(ipbuf);
			if (ip == (unsigned long int) UINT_MAX
					&& strcmp(ipbuf, BROADCAST)) {
				log(4, "Invalid IP: %s", ipbuf);
				return (struct hostlist_t *) 0;
			}
			if (*p == '/') {
				p++; i =0;
				while (!isspace((int)*p) && *p) {
					if (i < sizeof(ipbuf)-1) {
						ipbuf[i++] = *p;
					}
					p++;
				}
				ipbuf[i] = '\0';
				log(8, "With netmask %s", ipbuf);
				invalid = 0;

				if (strlen(ipbuf) < 3 && !strchr(ipbuf, '.')) {
					netmask = atoi(ipbuf);
					if (netmask < 0 || netmask > 32) {
						invalid = 1;
					}
					else {
						netmask = setlastbits(netmask);
					}
				} else {
					netmask = inet_addr(ipbuf);
					if (netmask == (unsigned long int)
								UINT_MAX
						&& strcmp(ipbuf, BROADCAST)) {
						invalid = 1;
					}
				}
				if (invalid) {
					log(4, "Invalid netmask: %s", ipbuf);

					/* Set the netmask to
					 * 255.255.255.255 */

					netmask = -1;
				}
			} else {
				/* no netmask specified */
				netmask = -1;
			}
			h_list = hostlist_ip_push(hl, ip, netmask);
		}
	}
	return h_list;
}


#define NAMESIZE  3
struct hostlist_t* config_parse_names(struct hostlist_t** hl,
							const char* line) {
	const char* p = line;
	char* buf, *tmp;
	char n[NAMESIZE];
	int i;
	size_t bufsize;

	struct hostlist_t* h_list = (struct hostlist_t*) 0;

	if ( !line ) { *hl = (struct hostlist_t*) '\0'; };

	for (;;) {
		while( p && *p && !isalpha((int)*p)) {
			/* see if it is a domain like .ibm.com */
			if (*p == '.') {
				if (p == line) {
					/* at the beginning of the line */
					break;
				}
				if (isspace((int)*(p-1))) {
					/* the dot is following a space chr */
					break;
				}
			}
			p++;
		}
		if ( !p || !*p ) {
			break;
		}
		buf = (char*) malloc(1);
		buf[0] = '\0';
		while(*p && !isspace((int)*p)) {
			i =0;
			while (i < NAMESIZE - 1 && !isspace((int)*p)) {
				n[i++] = *p++;
			}
			n[i] = '\0';
			tmp = buf;
			bufsize = strlen(buf) + sizeof(n) + 1;
			buf = (char*) malloc(bufsize);
			snprintf(buf, bufsize, "%s%s", tmp, n);
			free(tmp);
			tmp = 0;
		}
		log(9, "Read name %s", buf);

		h_list = hostlist_name_push(hl, buf);
	}
	return h_list;
}


void clean_hostlist(struct hostlist_t *hl) {
	if (!hl) {
		return;
	}
	if (hl->next) {
		clean_hostlist(hl->next);
		hl->next =0;
	}
	free(hl->host.name);
	free(hl);
}


unsigned int config_count_portrange(struct portrangestruct* prs) {
	if ( ! prs ) {
		return 0;
	}
	return (prs->endport - prs->startport + 1) +
		config_count_portrange(prs->next);
}

struct portrangestruct* portrangestruct_clone(struct portrangestruct* prs) {
	struct portrangestruct* new;
	if ( ! prs ) {
		return (struct portrangestruct*) 0;
	}

	new = (struct portrangestruct*) malloc(sizeof(struct portrangestruct));
	enough_mem(new);

	new->startport = prs->startport;
	new->endport = prs->endport;

	new->next = portrangestruct_clone(prs->next);

	return new;
}


struct portrangestruct* config_parse_portranges(const char* line) {
	int offset = 0;
	struct portrangestruct* prsbase, *prscur, *prs;
	char* startportstr, *endportstr;
	unsigned int startport, endport;

	struct slist_t* slistbase, *slistcur, *prevtmp;

	if ( ! line ) {
		return (struct portrangestruct*) 0;
	}
	slistcur = slistbase = config_split_line(line, WHITESPACES"\n"); 
	prsbase = prs = prscur = (struct portrangestruct*) 0;

	while (slistcur) {
		offset = 0;
		startportstr = quotstrtok(slistcur->value, ":", &offset);
		endportstr = quotstrtok(slistcur->value, ":", &offset);
		if (!startportstr) {
			log(3, "invalid port range: %s", slistcur->value);
			slist_destroy(slistbase);
			return (struct portrangestruct*) 0;
		}
		if (!endportstr) {
			size_t size = strlen(slistcur->value) * 2 + 1 + 1;
			char* newentry = (char*) malloc(size);
			snprintf(newentry, size, "%s:%s",
					slistcur->value, slistcur->value);
			free(slistcur->value);
			slistcur->value = newentry;
			offset = 0;
			startportstr = quotstrtok(slistcur->value, ":",
					&offset);
			endportstr = quotstrtok(slistcur->value, ":",
					&offset);
		}
		if (!*startportstr || !*endportstr) {
			log(3, "invalid port range: %s", slistcur->value);
			/* free */
			slist_destroy(slistbase);
			return (struct portrangestruct*) 0;
		}
		startport = strtol(startportstr, NULL, 10);
		if (errno == ERANGE
				&& (startport == LONG_MIN || startport == LONG_MAX)) {
			log(3, "Error reading the starting port number in %s",
				startportstr);
			slist_destroy(slistbase);
			free(startportstr);
			return (struct portrangestruct*) 0;
		}
		free(startportstr);
		startportstr = (char*) 0;

		endport = strtol(endportstr, NULL, 10);
		if (errno == ERANGE
				&& (endport == LONG_MIN || endport == LONG_MAX)) {
			log(3, "Error reading the ending port number in %s",
				endportstr);
			slist_destroy(slistbase);
			free(endportstr);
			return (struct portrangestruct*) 0;
		}
		free(endportstr);
		endportstr = (char*) 0;

		/* startport really below or equal to endport ? */
		if (startport > endport) {
			log(4, "Port range %d:%d invalid (starting port number"
				" %d is _above_ endport %d)",
					startport, endport,
					startport, endport);
			slist_destroy(slistbase);
			return (struct portrangestruct*) 0;
		}

		/* I want only to allow clear ranges. This is better for the
		 * checking, for the selection of a port and the admin is
		 * forced to specify his ports in a clear manner  :-)
		 *
		 * That means that a portrange like:
		 *
		 * 1000:2000   1500:1600   is invalid, since 1500:1600 is
		 * already in 1000:2000
		 *
		 * Also invalid:
		 *
		 * 1000:2000   1800:2200   (200 portnumbers are the same)
		 * 1000:2000   2000:3000   (1 portnumber is the same)
		 * 1000:2000   500:1000    (1 portnumber is the same)
		 * 1000:2000   799:1200    (200 portnumbers are the same)
		 *
		 * criteria: startport and endport => let's say just pno
		 *
		 * If pno is between (or equal to) the range of another port
		 * specification, state that is invalid.
		 * ------
		 * What happens if we first specify
		 *
		 * 300:400
		 *
		 * and then
		 *
		 * 200:1000 ?
		 *
		 * We have to do the checking for _each_ new range to _all_
		 * ranges. Compare every range to every other range
		 *
		 */

		prs = prsbase;
		while (prs) {
			if (
			    (startport >= prs->startport &&
			     startport <= prs->endport)
			    ||
			    (endport >= prs->startport &&
			     endport <= prs->endport)
			   ) {
				log(4, "Port range %d:%d invalid (parts "
					"covered by another range (%d:%d))",
						startport, endport,
						prs->startport, prs->endport);
				slist_destroy(slistbase);
				return (struct portrangestruct*) 0;
			}

			if (
			    (prs->startport >= startport &&
			     prs->startport <= endport)
			    ||
			    (prs->endport >= startport &&
			     prs->endport <= endport)
			    ) {
				log(4, "Port range %d:%d invalid (parts "
					"covered by another range (%d:%d))",
						prs->startport, prs->endport,
						startport, endport);
				slist_destroy(slistbase);
				return (struct portrangestruct*) 0;
			}
			prs = prs->next;
		}

		prs = (struct portrangestruct*)
				malloc(sizeof(struct portrangestruct));
		if (prscur) {
			prscur->next = prs;
		}

		prscur = prs;

		if (!prsbase) {
			/* the first one */
			prsbase = prscur;
		}
		prscur->startport = startport;
		prscur->endport = endport;
		prscur->next = (struct portrangestruct*) 0;

		/* iterate through and always chop off the first element */
		prevtmp = slistcur;
		slistcur = slistcur->next;
		slistbase = slistcur;
		free(prevtmp->value);
		free(prevtmp);
		prevtmp = 0;
	}

	return prsbase;
}


